@@ -1,7 +1,7 @@
 ;;; isl.el --- Simple incremental search in current-buffer -*- lexical-binding: t -*-
 
 ;; Author:      Thierry Volpiatto <thievol@posteo.net>
-;; Copyright (C) 2021~2024 Thierry Volpiatto <thievol@posteo.net>
+;; Copyright (C) 2021~2025 Thierry Volpiatto <thievol@posteo.net>
 
 ;; Version: 1.4
 ;; URL: https://github.com/thierryvolpiatto/isearch-light
@@ -38,11 +38,6 @@
   (require 'cl-lib)
   (require 'oclosure))
 
-;; Compatibility
-(unless (and (fboundp 'pos-bol) (fboundp 'pos-eol))
-  (defalias 'pos-bol 'line-beginning-position)
-  (defalias 'pos-eol 'line-end-position))
-
 (defvar iedit-aborting)
 (defvar iedit-read-only-occurrences-overlays)
 (defvar iedit-read-only-occurrences-overlays)
@@ -64,45 +59,13 @@
 (declare-function hs-show-block "hideshow.el")
 (declare-function markdown-show-entry "ext:markdown-mode.el")
 (declare-function bm-toggle "ext:bm.el")
+(declare-function pulse-momentary-highlight-region "pulse.el")
 
 ;; History
 (defvar isl-history nil)
 
-;; Internals
-(defvar isl-pattern "")
-(defvar-local isl-last-query nil)
-(put 'isl-last-query 'permanent-local t)
-(defvar-local isl-last-object nil)
-(put 'isl-last-object 'permanent-local t)
-(defvar isl-visited-buffers nil)
-(defvar isl-current-buffer nil)
-(defvar isl--item-overlays nil)
-(defvar isl--iterator nil)
-(defvar isl--last-overlay nil)
-(defvar isl--initial-pos nil)
-(defvar isl--number-results 0)
-(defvar isl--yank-point nil)
-(defvar isl--quit nil)
-(defvar isl--invalid nil)
-(defvar isl--window-start nil)
-(defvar-local isl--buffer-invisibility-spec nil)
-(defconst isl-space-regexp "\\s\\\\s-"
-  "Match a quoted space in a string.")
-(defconst isl--case-fold-choices '(smart nil t))
-(defvar isl--case-fold-choices-iterator nil)
-(defvar isl--hidding nil)
-(defvar isl--point-min nil)
-(defvar isl--point-max nil)
-(defvar isl--narrow-to-region nil)
-(defvar isl-search-invisible t)
-(defvar isl--buffer-tick nil)
 
-;; User vars
-
-(defvar isl-help-buffer-name "*isl help*")
-
-(defvar isl-timer-delay 0.01)
-
+;; Help
 (defvar isl-help-string
   "* ISL help\n
 
@@ -135,12 +98,50 @@ e.g. \"foo !bar\" would match any symbol containing foo but not bar.
 \\[isl-jump-to-helm-occur]\t\tJump to helm-occur
 \\[isl-jump-to-iedit-mode]\t\tJump to iedit-mode
 \\[isl-query-replace]\t\tJump to query replace
-\\[isl-show-or-hide-context-lines]\t\tHide or show non matching lines
+\\[isl-show-or-hide-context-lines]\t\tHide or show non matching lines (num prefarg or default)
 \\[isl-toggle-multi-search-in-line]\t\tToggle multi search style (InLine/InSymbol)
-\\[isl-toggle-invisible-search]\t\tToggle searching in invisible text
+\\[isl-toggle-invisible-search]\t\tToggle searching in invisible text (Sinv)
 \\[isl-bm-this-pos]\t\tAdd bookmark BM to current pos
-\\[isl-kill-selection]\t\tKill selected occurence
-\\[isl-align-regexp]\t\tAlign text matching regexp in region")
+\\[isl-kill-selection]\t\tKill selected occurence")
+
+
+;; Private internals
+(defvar isl--pattern "")
+(defvar-local isl--last-query nil)
+(put 'isl--last-query 'permanent-local t)
+(defvar-local isl--last-object nil)
+(put 'isl--last-object 'permanent-local t)
+(defvar isl--visited-buffers nil)
+(defvar isl--current-buffer nil)
+(defvar isl--item-overlays nil)
+(defvar isl--iterator nil)
+(defvar isl--last-overlay nil)
+(defvar isl--initial-pos nil)
+(defvar isl--number-results 0)
+(defvar isl--yank-point nil)
+(defvar isl--quit nil)
+(defvar isl--invalid nil)
+(defvar isl--window-start nil)
+(defvar-local isl--buffer-invisibility-spec nil)
+(defconst isl--space-regexp "\\s\\\\s-"
+  "Match a quoted space in a string.")
+(defconst isl--case-fold-choices '(smart nil t))
+(defvar isl--case-fold-choices-iterator nil)
+(defvar isl--hidding nil)
+(defvar isl--point-min nil)
+(defvar isl--point-max nil)
+(defvar isl--narrow-to-region nil)
+(defvar isl--buffer-tick nil)
+(defvar isl--closest-overlay nil)
+
+;; Others
+(defvar isl-help-buffer-name "*isl help*")
+(defvar isl-timer-delay 0.01)
+(defvar isl-search-invisible t)
+(defvar isl-prog-modes '(prog-mode))
+
+
+;; User vars
 
 (defgroup isl nil
   "Search buffers with `isl-search'."
@@ -191,9 +192,17 @@ in pattern."
   "Start updating after this number of chars."
   :type 'integer)
 
-(defcustom isl-visible-context-lines 1
-  "Number of lines to show around line when hiding non matching lines."
-  :type 'integer)
+(defcustom isl-visible-context-lines 0
+  "Number of lines to show around matched line when hiding context.
+Can be a cons cell like (nlines_before . nlines_after) or an integer,
+in this case it is used for before and after.
+When 0, the default show no context lines."
+  :type '(choice
+          (cons
+           :tag "Context lines number before and after"
+           (integer :tag "Show num lines before")
+           (integer :tag "Show num lines after"))
+          (integer :tag "Context lines number")))
 
 (defcustom isl-noresume-buffers '("*Helm Help*")
   "Prevent resuming in these buffers."
@@ -218,17 +227,13 @@ You can toggle this at any time with \\<isl-map>\\[isl-toggle-multi-search-in-li
 (defcustom isl-no-invisible-search-in-modes nil
   "Do not search invisible text in these modes."
   :type '(repeat symbol))
-
-(defcustom isl-align-regexp-group-pattern t
-  "Always surround isl-pattern with parens when non nil."
-  :type 'boolean)
 
 (defface isl-match
-  '((t :background "Brown4"))
+  '((t :background "brown4"))
   "Face used to highlight the items matched.")
 
 (defface isl-match-items
-  '((t :background "SaddleBrown" :foreground "black"))
+  '((t :background "OrangeRed3" :foreground "black"))
   "Face used to highlight the items matched inside line.")
 
 (defface isl-on
@@ -236,16 +241,12 @@ You can toggle this at any time with \\<isl-map>\\[isl-toggle-multi-search-in-li
        :foreground "black"))
   "Face used to highlight the item where point is.")
 
-(defface isl-line
-  '((t :background "Darkgoldenrod1" :extend t))
-  "Face used to flash line on exit.")
-
 (defface isl-number
   '((t :foreground "green"))
   "Face used to highlight number in mode-line.")
 
 (defface isl-string
-  '((t :foreground "Lightgoldenrod1" :bold t))
+  '((t :foreground "LightGoldenrod1" :bold t))
   "Face used to highlight pattern in mode-line.")
 
 (defface isl-case-fold
@@ -283,7 +284,6 @@ You can toggle this at any time with \\<isl-map>\\[isl-toggle-multi-search-in-li
     (define-key map (kbd "C-j")     'isl-toggle-multi-search-in-line)
     (define-key map (kbd "C-!")     'isl-bm-this-pos)
     (define-key map (kbd "C-c C-k") 'isl-kill-selection)
-    (define-key map (kbd "C-}")     'isl-align-regexp)
     map)
   "The map used when `isl-search' is running.
 Don't forget to modify `isl-mini-map' accordingly to fit with kmacros
@@ -333,12 +333,14 @@ Need to have the `bm' package installed."
 It put overlay on current position, move to next overlay using
 `isl--iterator', set `isl--yank-point' and then setup mode-line."
   (with-selected-window (minibuffer-selected-window)
-    (isl--highlight-last-overlay 'isl-match)
-    (when isl--iterator
-      ;; This is a noop when ARG==1 i.e. (1- 1) == 0.
-      (cl-loop repeat (1- arg) do (isl-iter-next isl--iterator))
-      (isl--goto-overlay (isl-iter-next isl--iterator)))
-    (isl-setup-mode-line)))
+    (if (cdr isl--item-overlays)
+        (progn
+          (isl--highlight-last-overlay 'isl-match)
+          (when isl--iterator
+            (cl-loop with ov repeat arg do (setq ov (isl-iter-next isl--iterator))
+                     finally (isl--goto-overlay ov)))
+          (isl-setup-mode-line))
+      (user-error "No more matches"))))
 
 (defun isl-scroll-1 (arg)
   "Scroll up if ARG is positive, down if it is negative."
@@ -352,14 +354,14 @@ It put overlay on current position, move to next overlay using
       (isl--find-and-goto-overlay ov))))
 
 (defun isl--first-ov-after-pos (pos)
-  "Move to next overlay from POS."
-  (cl-loop for ov in isl--item-overlays
+  "Find next overlay after POS."
+  (cl-loop for ov in (memql isl--last-overlay isl--item-overlays)
            when (> (overlay-start ov) pos)
            return ov))
 
 (defun isl--first-ov-before-pos (pos)
-  "Move to previous overlay from POS."
-  (cl-loop for ov in (reverse isl--item-overlays)
+  "Find previous overlay before POS."
+  (cl-loop for ov in (memql isl--last-overlay (reverse isl--item-overlays))
            when (< (overlay-start ov) pos)
            return ov))
 
@@ -376,12 +378,15 @@ It put overlay on current position, move to next overlay using
 (put 'isl-scroll-down 'no-helm-mx t)
 
 (defun isl-delete-minibuffer-contents ()
-  "No docstring."
+  "Delete minibuffer contents from isl prompt.
+If cursor is at eol delete whole contents otherwise delete from point
+to end."
   (interactive)
   (with-selected-window (minibuffer-window)
     (if (eolp)
         (delete-region (minibuffer-prompt-end) (point))
-      (delete-region (point) (point-max)))))
+      (delete-region (point) (point-max)))
+    (setq isl--yank-point nil)))
 (put 'isl-delete-minibuffer-contents 'no-helm-mx t)
 
 (defun isl--find-and-goto-overlay (overlay)
@@ -416,11 +421,9 @@ It put overlay on current position, move to next overlay using
 (defun isl-goto-closest-from-start ()
   "Goto closest match from start."
   (interactive)
-  (let ((ov (isl-closest-overlay
-             isl--initial-pos isl--item-overlays)))
-    (if (eql ov isl--last-overlay)
-        (user-error "Already at closest occurence from start")
-      (isl--find-and-goto-overlay ov))))
+  (if (eql isl--closest-overlay isl--last-overlay)
+      (user-error "Already at closest occurence from start")
+    (isl--find-and-goto-overlay isl--closest-overlay)))
 (put 'isl-goto-closest-from-start 'no-helm-mx t)
 
 (defun isl-goto-next (&optional arg)
@@ -461,10 +464,7 @@ It put overlay on current position, move to next overlay using
       (when isl-save-pos-to-mark-ring
         (set-marker (mark-marker) isl--initial-pos)
         (push-mark isl--initial-pos 'nomsg))
-      (let ((ov (make-overlay (pos-bol) (1+ (pos-eol)))))
-        (overlay-put ov 'face 'isl-line)
-        (sit-for 0.2)
-        (delete-overlay ov))
+      (pulse-momentary-highlight-region (pos-bol) (pos-eol))
       ;; Save `window-start' position before exiting minibuffer and
       ;; restore it after in `isl-cleanup' to avoid a moving screen.
       (setq isl--window-start (window-start))))
@@ -479,7 +479,7 @@ The word at point is relative to the current position in buffer, not
 the initial position i.e. the position before launching `isl-search'."
   (interactive)
   (let (str)
-    (with-current-buffer isl-current-buffer
+    (with-current-buffer isl--current-buffer
       (when (eolp) (error "End of line"))
       (when (or (memq (char-syntax (or (char-after) 0)) '(?w ?_ ? ))
                 (memq (char-syntax (or (char-after (1+ (point))) 0))
@@ -500,7 +500,7 @@ The symbol at point is relative to the current position in buffer, not
 the initial position i.e. the position before launching `isl-search'."
   (interactive)
   (let (str)
-    (with-current-buffer isl-current-buffer
+    (with-current-buffer isl--current-buffer
       (when (setq str (thing-at-point 'symbol t))
         (with-selected-window (minibuffer-window)
           (delete-minibuffer-contents)
@@ -523,13 +523,13 @@ the initial position i.e. the position before launching `isl-search'."
 (defun isl-change-matching-style ()
   "Toggle style matching in `isl-search' i.e. regexp/literal."
   (interactive)
-  (with-current-buffer isl-current-buffer
+  (with-current-buffer isl--current-buffer
     (setq-local isl-search-function
                 (cl-ecase isl-search-function
                   (isl--re-search-forward #'search-forward)
                   (search-forward #'isl--re-search-forward)))
     (unless executing-kbd-macro
-      (when (string= isl-pattern "")
+      (when (string= isl--pattern "")
         (let* ((style (isl-matching-style))
                (mode-line-format (format " Switching to %s searching" style)))
           (force-mode-line-update)
@@ -540,7 +540,7 @@ the initial position i.e. the position before launching `isl-search'."
 (defun isl-toggle-invisible-search ()
   "Toggle searching in invisible text."
   (interactive)
-  (with-current-buffer isl-current-buffer
+  (with-current-buffer isl--current-buffer
     (setq-local isl-search-invisible (not isl-search-invisible))
     (unless isl-search-invisible
       (setq buffer-invisibility-spec isl--buffer-invisibility-spec))
@@ -552,8 +552,8 @@ the initial position i.e. the position before launching `isl-search'."
   "Invoke `helm-occur' from `isl-search'."
   (interactive)
   (cl-assert (require 'helm-occur nil t) nil "Please install Helm package")
-  (let ((input isl-pattern)
-        (bufs (list isl-current-buffer)))
+  (let ((input isl--pattern)
+        (bufs (list isl--current-buffer)))
     (run-at-time 0.1 nil
                  (lambda ()
                    ;; Use `helm-occur-always-search-in-current' as a
@@ -571,7 +571,7 @@ automatically backward."
   (interactive "P")
   (cl-assert isl--item-overlays nil "Nothing yet to replace")
   (let* ((style    (isl-matching-style))
-         (regexp   isl-pattern)
+         (regexp   isl--pattern)
          (start    (overlay-start isl--last-overlay))
          (end      (overlay-end isl--last-overlay))
          (bottom   (overlay-start (car (last isl--item-overlays))))
@@ -592,7 +592,7 @@ automatically backward."
                       (format prompt (if wr " word" ""))
                       regexp-flag)
                      (and wr arg))))
-         (with-current-buffer isl-current-buffer
+         (with-current-buffer isl--current-buffer
            (save-excursion
              (let ((case-fold-search t))
                (goto-char (if backward end start))
@@ -650,9 +650,9 @@ Arguments OCCURRENCE-REGEXP, BEG and END have same meaning as in
   (cl-assert (require 'iedit nil t))
   (cl-assert isl--item-overlays nil "Nothing yet to replace")
   (let ((regexp (if (eq isl-search-function 'search-forward)
-                    (regexp-quote isl-pattern)
-                  isl-pattern))
-        (pos (with-current-buffer isl-current-buffer
+                    (regexp-quote isl--pattern)
+                  isl--pattern))
+        (pos (with-current-buffer isl--current-buffer
                (overlay-end isl--last-overlay))))
     (run-at-time
      0.1 nil
@@ -716,9 +716,14 @@ Arguments OCCURRENCE-REGEXP, BEG and END have same meaning as in
       (user-error "No help buffer found"))))
 (put 'isl-help-quit 'no-helm-mx t)
 
-(defun isl-show-or-hide-context-lines ()
-  "Hide or show non matching lines."
-  (interactive)
+(defun isl-show-or-hide-context-lines (&optional arg)
+  "Hide or show non matching lines.
+Show only numeric prefix ARG or `isl-visible-context-lines'
+number of lines around match.  Note that prefix arg allows specifying
+only the same number before and after matched lines whereas
+`isl-visible-context-lines' allows specifying a different value for
+before and after."
+  (interactive "p")
   (when isl--item-overlays
     (with-selected-window (minibuffer-selected-window)
       (if (setq isl--hidding (not isl--hidding))
@@ -730,7 +735,11 @@ Arguments OCCURRENCE-REGEXP, BEG and END have same meaning as in
               (set (make-local-variable 'line-move-ignore-invisible) t)
               (add-to-invisibility-spec '(isl-invisible . t))
               (while (not (eobp))
-                (forward-line (- isl-visible-context-lines))
+                (forward-line
+                 (- (or (and current-prefix-arg (max arg 0))
+                        (pcase isl-visible-context-lines
+                          (`(,before . ,_after) before)
+                          ((and same) same)))))
                 ;; Store position from n lines before
                 ;; this overlay and bol and move to next overlay.
                 (when (> (setq bol (pos-bol)) start)
@@ -738,7 +747,11 @@ Arguments OCCURRENCE-REGEXP, BEG and END have same meaning as in
                 (goto-char ov-end)
                 ;; Go to n lines after last overlay found and jump to
                 ;; next overlay from there.
-                (forward-line isl-visible-context-lines)
+                (forward-line
+                 (or (and current-prefix-arg (max arg 0))
+                     (pcase isl-visible-context-lines
+                       (`(,_before . ,after) after)
+                       ((and same) same))))
                 (setq start (1+ (pos-eol)))
                 (goto-char (next-single-char-property-change ov-end 'isl))
                 (setq ov-end (point)))
@@ -754,43 +767,6 @@ Arguments OCCURRENCE-REGEXP, BEG and END have same meaning as in
   (let ((ol (make-overlay beg end)))
     (overlay-put ol 'isl-invisible t)
     (overlay-put ol 'invisible 'isl-invisible)))
-
-(defun isl--maybe-revert-to-original ()
-  "Maybe restore original buffer contents on quit.
-Buffer may have been modified by `isl-align-regexp'."
-  (let ((contents isl--narrow-to-region))
-    (when (and (stringp contents)
-               (> (buffer-modified-tick) isl--buffer-tick))
-      ;; When isl--narrow-to-region is non nil isl--point-min and
-      ;; max are set as well.
-      (delete-region isl--point-min isl--point-max)
-      (insert contents))))
-
-(defun isl-align-regexp (arg)
-  "Align text matching regexp in current-buffer.
-Numeric prefix ARG is applied to the SPACING arg of `align-regexp'.
-Quitting undo all the align actions done in current session, to valid
-all align operations you have to exit with RET."
-  (interactive "p")
-  (cl-assert isl--item-overlays nil "Nothing yet to align")
-  (cl-assert isl--narrow-to-region nil "No region found")
-  (cl-assert (not (cdr (isl-split-string isl-pattern))) nil
-             "Can't align with a multi match expression")
-  (with-current-buffer isl-current-buffer
-    (align-regexp isl--point-min isl--point-max
-                  (concat "\\(\\s-+\\)"
-                          (if (equal (isl-matching-style) "Regex")
-                              (if isl-align-regexp-group-pattern
-                                  (format "\\(%s\\)" isl-pattern)
-                                isl-pattern)
-                            (regexp-quote isl-pattern)))
-                  1 arg)
-    ;; Align-regexp has probably added or removed spaces so update
-    ;; isl--point-min/max.
-    (setq isl--point-min (point-min)
-          isl--point-max (point-max))
-    (isl-update)))
-(put 'isl-align-regexp 'no-helm-mx t)
 
 ;;; Iterators
 ;;
@@ -815,9 +791,7 @@ The iterator is mutable, its direction can be changed at anytime with
 the function `isl-iterator-reverse', when doing so the sequence
 handled by the iterator is reversed from the last element yielded.
 Elements of the iterator can be accessed via the isl-iterator--* fns.
-All the oclosure slots are mutable but the only one that can be
-changed safely is \\='changing-direction', expect weird behavior if
-you try to modify other elements externally."
+All the oclosure slots are mutables."
   (oclosure-lambda (isl-iterator (seq seq)
                                  (tmp-seq seq)
                                  (element (car seq))
@@ -834,7 +808,7 @@ you try to modify other elements externally."
                             ('right 'left))
                 index (- (length seq) (length queue))
                 seq (append queue (take index revlist))
-                tmp-seq (cddr seq)
+                tmp-seq (or (cddr seq) seq) ; Fix seq with only two elements.
                 elm (cadr seq)
                 changing-direction nil)
         (setq tmp-seq (or (cdr tmp-seq) seq)))
@@ -867,10 +841,11 @@ If `isl--iterator' is non nil update it otherwise create it."
     (remove-overlays nil nil 'isl t)
     (setq isl--item-overlays nil)))
 
-(cl-defun isl-set-case-fold-search (&optional (pattern isl-pattern))
+(defun isl-set-case-fold-search (&optional pattern)
   "Return a suitable value for `case-fold-search'.
 This is done according to `isl-case-fold-search'.
-Optional argument PATTERN default to `isl-pattern'."
+Optional argument PATTERN default to `isl--pattern'."
+  (unless pattern (setq pattern isl--pattern))
   (cl-case isl-case-fold-search
     (smart (let ((case-fold-search nil))
              (if (string-match "[[:upper:]]" pattern) nil t)))
@@ -879,7 +854,7 @@ Optional argument PATTERN default to `isl-pattern'."
 (defun isl-select-case-fold-search ()
   "Set `case-fold-search' from `isl-search' session."
   (interactive)
-  (with-current-buffer isl-current-buffer
+  (with-current-buffer isl--current-buffer
     (if (eq last-command 'isl-select-case-fold-search)
         (setq-local isl-case-fold-search
                     (isl-iter-next isl--case-fold-choices-iterator))
@@ -897,7 +872,7 @@ Optional argument PATTERN default to `isl-pattern'."
   "Split string STR at non quoted spaces."
   (split-string
    (replace-regexp-in-string
-    isl-space-regexp "\\s-" str nil t)))
+    isl--space-regexp "\\s-" str nil t)))
 
 (defun isl-patterns (str)
   "Return an alist of (pred . regexp) elements from STR."
@@ -942,9 +917,7 @@ symbol or line position according to `isl-multi-search-in-line'."
                                  (cons (pos-bol) (1+ (pos-eol))))
                                 (rest
                                  (bounds-of-thing-at-point
-                                  (if (and (derived-mode-p 'prog-mode)
-                                           (not (nth 3 (syntax-ppss))))
-                                      'symbol 'filename)))
+                                  (isl--type-for-mode)))
                                 (t (cons (match-beginning 0) (match-end 0))))
              unless bounds return nil
              if (or (not rest)
@@ -968,7 +941,7 @@ symbol or line position according to `isl-multi-search-in-line'."
 (defun isl-toggle-multi-search-in-line ()
   "Toggle multi-search in lines/symbols."
   (interactive)
-  (with-current-buffer isl-current-buffer
+  (with-current-buffer isl--current-buffer
     (setq-local isl-multi-search-in-line (not isl-multi-search-in-line))
     (unless executing-kbd-macro
       (if isl-multi-search-in-line
@@ -982,7 +955,7 @@ See `isl-requires-pattern'."
   (> (length str) isl-requires-pattern))
 
 (defun isl-update ()
-  "Update `current-buffer' when `isl-pattern' change."
+  "Update `current-buffer' when `isl--pattern' change."
   (with-selected-window (minibuffer-selected-window)
     (while-no-input
       (when isl--hidding
@@ -993,7 +966,7 @@ See `isl-requires-pattern'."
       (setq isl--invalid nil)
       ;; We don't use the isearch-invisible mechanism which is heavy
       ;; and don't behave as we want, instead remove invisibility in
-      ;; all buffer and on exit restore it and unhide only the place
+      ;; all buffers and on exit restore it and unhide only the place
       ;; where point is with appropriate functions belonging to
       ;; major-mode e.g. org => org-reveal etc...
       (when (and isl-search-invisible
@@ -1001,14 +974,14 @@ See `isl-requires-pattern'."
                  (listp buffer-invisibility-spec))
         (mapc 'remove-from-invisibility-spec buffer-invisibility-spec))
       (let ((count 1)
-            (npos (point-max)) ; nearest match position from start.
-            ov bounds go ovs-alist)
-        (unless (string= isl-pattern "")
+            (npos (point-max)) ; start comparing at highest pos [1].
+            ov bounds go nearest)
+        (unless (string= isl--pattern "")
           (save-excursion
             (goto-char (point-min))
             (condition-case-unless-debug nil
                 (while (and (setq bounds (or go (isl-multi-search-fwd
-                                                 isl-pattern nil t)))
+                                                 isl--pattern nil t)))
                             (not (eobp)))
                   (setq go nil)
                   (unless (and (not isl-search-invisible)
@@ -1032,7 +1005,9 @@ See `isl-requires-pattern'."
                            (ovpos (car bounds))
                            (diff  (if (> pos ovpos)
                                       (- pos ovpos) (- ovpos pos))))
-                      (push (cons diff ov) ovs-alist)
+                      ;; [1] At first run, npos is (point-max) so diff is
+                      ;; always smaller at this time.
+                      (when (< diff npos) (setq nearest ov))
                       (setq npos (min diff npos)))
                     (when isl-multi-search-in-line
                       (isl--highlight-items-in-line (car bounds) (cdr bounds)))
@@ -1043,24 +1018,26 @@ See `isl-requires-pattern'."
                     ;; `isl--re-search-forward' will be stuck on next
                     ;; turn of the loop so match it with looking-at
                     ;; and use GO value in next turn of the loop.
-                    (when (looking-at isl-pattern)
+                    (when (looking-at isl--pattern)
                       (setq go (cons (pos-bol) (pos-eol))))))
               (invalid-regexp (setq isl--invalid t) nil))
             (setq isl--item-overlays (nreverse isl--item-overlays)))
           (if (null isl--item-overlays)
               (progn (setq isl--number-results 0)
                      (and isl--initial-pos (goto-char isl--initial-pos)))
-            (setq isl--last-overlay (cdr (assq npos ovs-alist))
-                  isl--number-results (max (length isl--item-overlays) 0))
+            (setq isl--last-overlay    nearest ; Will change while moving.
+                  isl--closest-overlay nearest ; Will change only on next update.
+                  isl--number-results  (max (length isl--item-overlays) 0))
             (isl--highlight-last-overlay 'isl-on)
             (isl-make-or-update-iterator)
-            (goto-char (overlay-end (isl-iter-next isl--iterator)))
+            (let ((next (isl-iter-next isl--iterator)))
+              (when next (goto-char (overlay-end next))))
             (setq isl--yank-point (point)))))
       (isl-setup-mode-line))))
 
 (defun isl--highlight-items-in-line (beg end)
   "Highlight items inside a matched line from BEG to END."
-  (let ((split (isl-split-string isl-pattern)))
+  (let ((split (isl-split-string isl--pattern)))
     ;; Do nothing until we have a multi pattern.
     (when (cdr split)
       ;; When this is called we are at eol.
@@ -1070,7 +1047,7 @@ See `isl-requires-pattern'."
                  for p in split
                  unless (string-match "\\`!" p)
                  do (save-excursion
-                      (while (isl--re-search-forward p end t)
+                      (while (funcall isl-search-function p end t)
                         (setq ov2 (make-overlay (match-beginning 0) (match-end 0)))
                         (overlay-put ov2 'face 'isl-match-items)
                         (overlay-put ov2 'isl t)
@@ -1083,7 +1060,7 @@ See `isl-requires-pattern'."
         (search (if isl-multi-search-in-line
                     (propertize "Inline" 'help-echo "Search in line")
                   (propertize "Insym" 'help-echo "Search in symbol")))
-        (position (with-current-buffer isl-current-buffer
+        (position (with-current-buffer isl--current-buffer
                      (if (and isl--initial-pos
                               (> (point) isl--initial-pos))
                          isl-after-position-string
@@ -1094,8 +1071,8 @@ See `isl-requires-pattern'."
                        isl-direction-up-string))))
     (when (numberp isl--number-results)
       (setq mode-line-format
-            (cond ((or (string= isl-pattern "")
-                       (<= (length isl-pattern)
+            (cond ((or (string= isl--pattern "")
+                       (<= (length isl--pattern)
                            isl-requires-pattern))
                    (default-value 'mode-line-format))
                   ((zerop isl--number-results)
@@ -1107,7 +1084,7 @@ See `isl-requires-pattern'."
                                                       isl-warning-char)
                                               'face 'font-lock-warning-face)
                                            "No results found for pattern")
-                                         (propertize isl-pattern
+                                         (propertize isl--pattern
                                                      'face 'isl-string)
                                          style
                                          search
@@ -1145,23 +1122,14 @@ See `isl-requires-pattern'."
                                                 'help-echo "case-fold-search")))
                            " " mode-line-position)))))))
 
-(defun isl-closest-overlay (pos overlays)
-  "Return closest overlay from POS in OVERLAYS list."
-  (cl-loop for ov in overlays
-           for ovpos = (overlay-start ov)
-           for diff = (if (> pos ovpos) (- pos ovpos) (- ovpos pos))
-           collect (cons diff ov) into res
-           minimize diff into min
-           finally return (cdr (assq min res))))
-
 (defun isl-check-input ()
   "Check minibuffer input."
   (with-selected-window (minibuffer-window)
     (let ((input (minibuffer-contents)))
-      (when (not (string= input isl-pattern))
-        (setq isl-pattern input)
-        (if (and (stringp isl-pattern)
-                 (isl-maybe-update isl-pattern))
+      (when (not (string= input isl--pattern))
+        (setq isl--pattern input)
+        (if (and (stringp isl--pattern)
+                 (isl-maybe-update isl--pattern))
             (isl-update)
           (with-selected-window (minibuffer-selected-window)
             (isl-delete-overlays)
@@ -1184,16 +1152,16 @@ Arguments INITIAL-INPUT and DEFAULT are same as in `read-from-minibuffer'."
 
 (defun isl-cleanup ()
   "Cleanup various things when `isl-search' exit."
-  (with-current-buffer isl-current-buffer
+  (with-current-buffer isl--current-buffer
     (let* ((pos (and isl--last-overlay ; nil when quitting with no results.
                      (overlay-end isl--last-overlay)))
            (hs-show-hook (list (lambda () (and pos (goto-char pos))))))
       (when (buffer-live-p (get-buffer isl-help-buffer-name))
         (kill-buffer isl-help-buffer-name))
-      (setq isl-last-object
+      (setq isl--last-object
             `(lambda ()
                (setq-local mode-line-format ',mode-line-format
-                           isl-last-query ,isl-pattern
+                           isl--last-query ,isl--pattern
                            isl--initial-pos ,pos
                            isl--point-min ,isl--point-min
                            isl--point-max ,isl--point-max
@@ -1206,12 +1174,13 @@ Arguments INITIAL-INPUT and DEFAULT are same as in `read-from-minibuffer'."
                            isl--hidding ,isl--hidding
                            isl-search-invisible ,isl-search-invisible
                            isl-multi-search-in-line ,isl-multi-search-in-line
-                           cursor-in-non-selected-windows ,cursor-in-non-selected-windows)))
+                           cursor-in-non-selected-windows ',cursor-in-non-selected-windows)))
       (isl-delete-overlays)
       (setq mode-line-format (default-value 'mode-line-format)
             isl--yank-point nil
             isl--iterator nil
             isl--item-overlays nil
+            isl--closest-overlay nil
             isl--last-overlay nil
             isl--number-results nil
             isl-case-fold-search (default-value 'isl-case-fold-search)
@@ -1254,18 +1223,18 @@ Note that INPUT cannot be used with a non nil value for RESUME."
              nil "(isl-search-1): RESUME and INPUT can't be used together")
   (unless resume
     (setq isl--initial-pos (point)
-          isl-pattern ""
-          isl-current-buffer (current-buffer)
+          isl--pattern ""
+          isl--current-buffer (current-buffer)
           isl--buffer-tick (buffer-modified-tick)
           isl--buffer-invisibility-spec buffer-invisibility-spec
-          cursor-in-non-selected-windows nil))
+          cursor-in-non-selected-windows 'bar))
   (setq isl--window-start nil)
-  (when (and (buffer-live-p isl-current-buffer)
-             (not (member (buffer-name isl-current-buffer)
+  (when (and (buffer-live-p isl--current-buffer)
+             (not (member (buffer-name isl--current-buffer)
                           isl-noresume-buffers)))
-    (setq isl-visited-buffers
-          (cons isl-current-buffer
-                (delete isl-current-buffer isl-visited-buffers))))
+    (setq isl--visited-buffers
+          (cons isl--current-buffer
+                (delete isl--current-buffer isl--visited-buffers))))
   (when (memq major-mode isl-no-invisible-search-in-modes)
     (setq-local isl-search-invisible nil))
   (let* ((format-fn (if (eq isl-search-function 'search-forward)
@@ -1290,27 +1259,32 @@ Note that INPUT cannot be used with a non nil value for RESUME."
               "Search: "
               (if resume
                   (buffer-local-value
-                   'isl-last-query isl-current-buffer)
+                   'isl--last-query isl--current-buffer)
                 input)
               default)
            (quit
             (setq isl--quit t)
-            (isl--maybe-revert-to-original)
             (when isl--initial-pos
               (goto-char isl--initial-pos))))
       (isl-cleanup)
       ;; Avoid loosing focus in helm help buffer.
       (unless (eq (window-buffer (selected-window))
-                  isl-current-buffer)
-        (switch-to-buffer isl-current-buffer)))))
+                  isl--current-buffer)
+        (switch-to-buffer isl--current-buffer)))))
+
+(defun isl--type-for-mode ()
+  "Return the type of thing to use with `thing-at-point' according to context."
+  (if (and (derived-mode-p isl-prog-modes)
+           (not (nth 3 (syntax-ppss))))
+      'symbol 'filename))
 
 (defun isl--thing-at-point ()
-  "Return symbol or filename at point."
-  (thing-at-point
-   (if (and (derived-mode-p 'prog-mode)
-            (not (nth 3 (syntax-ppss))))
-       'symbol 'filename)
-   t))
+  "Collect symbol or filename at point in all visible buffers.
+The return value is a list."
+  (cl-loop for win in (window-list nil 1)
+           when (with-selected-window win
+                  (thing-at-point (isl--type-for-mode) t))
+           collect it))
 
 (defvar isl-mini-map
   (let ((map (make-sparse-keymap)))
@@ -1329,7 +1303,7 @@ modifying keybindings here.")
 This function is intended to be used in kmacros."
   (unwind-protect
        ;; Needed when toggling inline search.
-       (let* ((isl-current-buffer (current-buffer))
+       (let* ((isl--current-buffer (current-buffer))
               (str (read-from-minibuffer "Search: " nil isl-mini-map)))
          (isl-multi-search-fwd str nil t))
     ;; Reset local var to their default value for next iteration.
@@ -1356,23 +1330,23 @@ stop, assuming user starts its macro above the text to edit."
   "Resume isl session in current buffer.
 With a prefix ARG choose one of the last buffers isl had visited."
   (interactive "P")
-  (setq isl-current-buffer
-        (cond ((and arg isl-visited-buffers)
+  (setq isl--current-buffer
+        (cond ((and arg isl--visited-buffers)
                (get-buffer
                 (completing-read
                  "Resume from buffer: "
-                 (delq nil (mapcar 'buffer-name isl-visited-buffers))
+                 (delq nil (mapcar 'buffer-name isl--visited-buffers))
                  nil t)))
-              (isl-visited-buffers
+              (isl--visited-buffers
                (car (memql (current-buffer)
-                           isl-visited-buffers)))))
-  (cl-assert isl-current-buffer
+                           isl--visited-buffers)))))
+  (cl-assert isl--current-buffer
              nil "No previous Isl session yet recorded here")
-  (switch-to-buffer isl-current-buffer)
+  (switch-to-buffer isl--current-buffer)
   (let (beg end)
-    (with-current-buffer isl-current-buffer
-      (funcall isl-last-object)
-      (setq isl-pattern "")
+    (with-current-buffer isl--current-buffer
+      (funcall isl--last-object)
+      (setq isl--pattern "")
       ;; Last session was isl-narrow-to-defun.
       (setq beg isl--point-min end isl--point-max))
     (save-restriction
@@ -1382,7 +1356,7 @@ With a prefix ARG choose one of the last buffers isl had visited."
 
 ;;;###autoload
 (defun isl-narrow-to-region-or-defun (beg end)
-  "Start incremental searching in region or current defun."
+  "Start incremental searching in region from BEG to END or current defun."
   (interactive "r")
   (setq isl--point-min nil
         isl--point-max nil
